# See LICENSE.incore for license details
import logging
import os
import sys
import shutil
from datetime import datetime


from repomanager import __version__ as version
import repomanager.utils as utils
import repomanager.rpm as repoman

def main():
    '''
        Entry point for riscv_config.
    '''

    # Set up the parser
    parser = utils.repomanager_cmdline_args()
    args = parser.parse_args()

    # Set up the logger
    utils.setup_logging(args.verbose)
    logger = logging.getLogger()
    logger.handlers = []
    ch = logging.StreamHandler()
    ch.setFormatter(utils.ColoredFormatter())
    logger.addHandler(ch)

    logger.info('################### Repository Manager ##################')
    logger.info(f'--- Copyright (c) 2020-{datetime.now().year} InCore Semiconductors Pvt. Ld. ---')
    logger.info('------------------- Version : {0} ---------------------'.format(version))
    logger.info('#########################################################')
    logger.info('\n')

    yml_str = open(args.yaml,'r').read()
    repoman.repoman(yml_str, args.clean, args.update, args.patch,
            args.unpatch, args.dir)

if __name__ == "__main__":
    exit(main())



# See LICENSE for details

from click.testing import CliRunner
from repomanager.main import cli
import pytest 

@pytest.fixture
def runner():
    return CliRunner()

def test_clean(runner):
    '''Testing clean option'''
    result = runner.invoke(cli, ['--clean'])
    assert result.exit_code == 0

def test_version(runner):
    '''Testing version option'''
    result = runner.invoke(cli, ['--version'])
    assert result.exit_code == 0

def test_clean_update_patch(runner):
    result = runner.invoke(cli, ['-cup'])
    assert result.exit_code == 0
    result = runner.invoke(cli, ['-c'])
    assert result.exit_code == 0

def test_repolist(runner):
    result = runner.invoke(cli, ['--repolist', 'repolist1.yaml', '-cup'])
    assert result.exit_code == 0
    result = runner.invoke(cli, ['--repolist', 'repolist1.yaml', '-c'])
    assert result.exit_code == 0

def test_master(runner):
    result = runner.invoke(cli, ['--repolist', 'repo_master.yaml', '-cup'])
    assert result.exit_code == 0
    result = runner.invoke(cli, ['--repolist', 'repo_master.yaml', '-c'])
    assert result.exit_code == 0

def test_nopatch(runner):
    result = runner.invoke(cli, ['--repolist', 'repo_nopatch.yaml', '-cup'])
    assert result.exit_code == 0
    result = runner.invoke(cli, ['--repolist', 'repo_nopatch.yaml', '-c'])
    assert result.exit_code == 0

#def test_submodule(runner):
#    result = runner.invoke(cli, ['--repolist', 'repo_submodule.yaml', '-r', '--verbose', 'info'])
#    assert result.exit_code == 0
#    result = runner.invoke(cli, ['--repolist', 'repo_submodule.yaml', '-p'])
#    assert result.exit_code == 0
#    result = runner.invoke(cli, ['--repolist', 'repo_submodule.yaml', '--unpatch'])
#    assert result.exit_code == 0
#    result = runner.invoke(cli, ['--repolist', 'repo_submodule.yaml', '--unpatch'])
#    assert result.exit_code == 0
#    result = runner.invoke(cli, ['--repolist', 'repo_submodule.yaml', '-p'])
#    assert result.exit_code == 0

def test_error(runner):
    result = runner.invoke(cli, ['--repolist', 'error.yaml', '-cup'])
    assert result.exit_code != 0

def test_dir(runner):
    '''Testing dir option'''
    result = runner.invoke(cli, ['--dir', 'work', '-cup'])
    assert result.exit_code == 0




# See LICENSE for details

import logging

logger = logging.getLogger(__name__)

class ColoredFormatter(logging.Formatter):
    """                                                                         
        Class to create a log output which is colored based on level.           
    """

    def __init__(self, *args, **kwargs):
        super(ColoredFormatter, self).__init__(*args, **kwargs)
        self.colors = {
            'DEBUG': '\033[94m',
            'INFO': '\033[92m',
            'WARNING': '\033[93m',
            'ERROR': '\033[91m',
        }

        self.reset = '\033[0m'

    def format(self, record):
        msg = str(record.msg)
        level_name = str(record.levelname)
        name = str(record.name)
        color_prefix = self.colors[level_name]
        return '{0}{1:<9s} : {2}{3}'.format(color_prefix,
                                            '[' + level_name + ']', msg,
                                            self.reset)


def setup_logging(log_level):
    """Setup logging                                                            
                                                                                
        Verbosity decided on user input                                         
                                                                                                                                                   
        :param log_level: User defined log level                             
                                                                                
        :type log_level: str                                                               
    """
    numeric_level = getattr(logging, log_level.upper(), None)

    if not isinstance(numeric_level, int):
        print(
            "\033[91mInvalid log level passed. Please select from debug | info | warning | error\033[0m"
        )
        raise ValueError("{}-Invalid log level.".format(log_level))

    logging.basicConfig(level=numeric_level)




# See LICENSE.incore for license details
import pathlib
import subprocess
import logging
import argparse
import operator
import shutil
import sys
import shlex
import ruamel
import time
from ruamel.yaml import YAML
import re

yaml = YAML(typ="rt")
yaml.default_flow_style = False
yaml.allow_unicode = True

logger = logging.getLogger(__name__)

def load_yaml(foo):
    try:
        with open(foo, "r") as file:
            return yaml.load(file)
    except ruamel.yaml.constructor.DuplicateKeyError as msg:
        logger = logging.getLogger(__name__)
        error = "\n".join(str(msg).split("\n")[2:-7])
        logger.error(error)
        raise SystemExit


class ColoredFormatter(logging.Formatter):
    """                                                                         
        Class to create a log output which is colored based on level.           
    """

    def __init__(self, *args, **kwargs):
        super(ColoredFormatter, self).__init__(*args, **kwargs)
        self.colors = {
            'DEBUG': '\033[94m',
            'INFO': '\033[92m',
            'WARNING': '\033[93m',
            'ERROR': '\033[91m',
        }

        self.reset = '\033[0m'

    def format(self, record):
        msg = str(record.msg)
        level_name = str(record.levelname)
        name = str(record.name)
        color_prefix = self.colors[level_name]
        funcName = str(record.funcName)
        if level_name == 'DEBUG':
            return '{0}{1:>9s} | {4} | {2}{3}'.format(color_prefix,
                                                level_name , msg,
                                                self.reset, name+'.'+funcName)
        else:
            return '{0}{1:>9s} | {2}{3}'.format(color_prefix,
                                                level_name , msg,
                                            self.reset)


def setup_logging(log_level):
    """Setup logging                                                            
                                                                                
        Verbosity decided on user input                                         
                                                                                                                                                   
        :param log_level: User defined log level                             
                                                                                
        :type log_level: str                                                               
    """
    numeric_level = getattr(logging, log_level.upper(), None)

    if not isinstance(numeric_level, int):
        print(
            "\033[91mInvalid log level passed. Please select from debug | info | warning | error\033[0m"
        )
        raise ValueError("{}-Invalid log level.".format(log_level))

    logging.basicConfig(level=numeric_level)


class SortingHelpFormatter(argparse.HelpFormatter):

    def add_arguments(self, actions):
        actions = sorted(actions, key=operator.attrgetter('option_strings'))
        super(SortingHelpFormatter, self).add_arguments(actions)


def repomanager_cmdline_args():
    parser = argparse.ArgumentParser(formatter_class=SortingHelpFormatter,
                                     prog="repomanager",
                                     description="Repository Manager")
    parser.add_argument('--dir','-d',
                        type=str,
                        metavar='PATH',
                        default='./',
                        help='PATH of the working directory')

    parser.add_argument('--clean','-c',
                        action='store_true',
                        default=False,
                        help='Removes all the repositories listed in repolist'
                        )
    parser.add_argument('--yaml','-y',
                        type=str,
                        metavar = 'YAML',
                        default = 'repolist.yaml',
                        help='A YAML file containing list of the repos' 
                        )
    parser.add_argument('--update','-u',
                        action='store_true',
                        default=False,
                        help='Updates/Clones all the listed repositories in the repolist'
                        )
    parser.add_argument('--patch','-p',
                        action='store_true',
                        default=False,
                        help='Applies patches to repositories which have their patch field defined in repolist'
                        )
    parser.add_argument('--unpatch','-r',
                        action='store_true',
                        default=False,
                        help='Removes patches to repositories which have their patch field defined in repolist'
                        )
    parser.add_argument('--verbose',
                        action='store',
                        default='info',
                        help='debug | info | warning | error',
                        metavar="")
    return parser

def which(program):
    x = shutil.which(program)
    if x is None:
        logger.error(program+' not found')
        sys.exit(0)
    else:
        return x

def find_packages(file_contents):
    file_lines = file_contents.split('\n')
    pattern = re.compile(r'import\s+(.*)\s*::')
    import_lines = filter(lambda x :
        pattern.search(x) != None, file_lines)
    packages = map(lambda x: 
        pattern.search(x).groups()[0].strip(), import_lines)

    return list(packages)

def recurse_packages(package, files, rules, done, prelude):
    # package file
    try:
        package_file = list(filter(lambda x : package.strip() + '.bsv' in x,
            files))[0]
    except IndexError as e:
        return

    with open(package_file, 'r') as f:
        contents = f.read()

    # build rule
    deps = list(filter(lambda x : not x in prelude, find_packages(contents)))
    deps_string = ', '.join(deps)

    rule = '''
{0} = custom_target('{0}',
    output : '{0}.bo',
    input : '{1}',
    command : [bsc] + options + ['@INPUT@'],
    depends : [{2}] )'''.format(
        package, package_file, deps_string)

    rules += [(len(deps), rule)]
    done += [package]

    for p in deps:
        if p in done:
            continue
        else:
            recurse_packages(p, files, rules, done, prelude)
    return list(rules)

class Command():
    """
    Class for command build which is supported
    by :py:mod:`suprocess` module. Supports automatic
    conversion of :py:class:`pathlib.Path` instances to
    valid format for :py:mod:`subprocess` functions.
    """

    def __init__(self, *args, pathstyle='auto', ensure_absolute_paths=False):
        """Constructor.

        :param pathstyle: Determine the path style when adding instance of
            :py:class:`pathlib.Path`. Path style determines the slash type
            which separates the path components. If pathstyle is `auto`, then
            on Windows backslashes are used and on Linux forward slashes are used.
            When backslashes should be prevented on all systems, the pathstyle
            should be `posix`. No other values are allowed.

        :param ensure_absolute_paths: If true, then any passed path will be
            converted to absolute path.

        :param args: Initial command.

        :type pathstyle: str

        :type ensure_absolute_paths: bool
        """
        self.ensure_absolute_paths = ensure_absolute_paths
        self.pathstyle = pathstyle
        self.args = []

        for arg in args:
            self.append(arg)

    def append(self, arg):
        """Add new argument to command.

        :param arg: Argument to be added. It may be list, tuple,
            :py:class:`Command` instance or any instance which
            supports :py:func:`str`.
        """
        to_add = []
        if type(arg) is list:
            to_add = arg
        elif type(arg) is tuple:
            to_add = list(arg)
        elif isinstance(arg, type(self)):
            to_add = arg.args
        elif isinstance(arg, str) and not self._is_shell_command():
            to_add = shlex.split(arg)
        else:
            # any object which will be converted into str.
            to_add.append(arg)

        # Convert all arguments to its string representation.
        # pathlib.Path instances
        to_add = [
            self._path2str(el) if isinstance(el, pathlib.Path) else str(el)
            for el in to_add
        ]
        self.args.extend(to_add)

    def clear(self):
        """Clear arguments."""
        self.args = []

    def run(self, logging=True, **kwargs):
        """Execute the current command.

        Uses :py:class:`subprocess.Popen` to execute the command.

        :return: The return code of the process     .
        :raise subprocess.CalledProcessError: If `check` is set
                to true in `kwargs` and the process returns
                non-zero value.
        """
        kwargs.setdefault('shell', self._is_shell_command())
        cwd = self._path2str(kwargs.get(
            'cwd')) if not kwargs.get('cwd') is None else self._path2str(
                os.getcwd())
        kwargs.update({'cwd': cwd})
        logger.debug(cwd)
        # When running as shell command, subprocess expects
        # The arguments to be string.
        logger.debug(str(self))
        cmd = str(self) if kwargs['shell'] else self
        out = "Max retries reached. Could not resolve host."
        xret_code = -1
        max_retries = 5
        retry_delay = 5
        for attempt in range(max_retries):
            try:
                x = subprocess.run(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE, encoding="utf-8", **kwargs)
                # x = subprocess.Popen(cmd, stdout=subprocess.PIPE,stderr=subprocess.PIPE,  **kwargs)
                xret_code = x.returncode
                out = x.stdout
                err = x.stderr
            except Exception as e:
                logger.warning(e)
                logger.error('Failed to clone "{0}" at directory "{1}". Reason : {2}'.format(cmd, cwd, str(e)))
                break
            else:
                if xret_code != 0 and logging:
            if out:
                        logger.warning(out)
                    if "could not resolve host" in err.lower():
                        logger.warning("Retrying connection to host . . .")
                        time.sleep(retry_delay)
                    else:
            if err:
                            logger.error(err)
                        break
        elif logging:
            if out:
                        logger.debug(out)
            if err:
                        logger.debug(err)
                    break
                else:
                    break
        else:
            logger.error("Max retries reached. Could not resolve host.")
            raise SystemExit(1)
        return out, xret_code

    def _is_shell_command(self):
        """
        Return true if current command is supposed to be executed
        as shell script otherwise false.
        """
        return any('|' in arg for arg in self.args)

    def _path2str(self, path):
        """Convert :py:class:`pathlib.Path` to string.

        The final form of the string is determined by the
        configuration of `Command` instance.

        :param path: Path-like object which will be converted
                     into string.
        :return: String representation of `path`
        """
        path = pathlib.Path(path)
        if self.ensure_absolute_paths and not path.is_absolute():
            path = path.resolve()

        if self.pathstyle == 'posix':
            return path.as_posix()
        elif self.pathstyle == 'auto':
            return str(path)
        else:
            raise ValueError(f"Invalid pathstyle {self.pathstyle}")

    def __add__(self, other):
        cmd = Command(self,
                      pathstyle=self.pathstyle,
                      ensure_absolute_paths=self.ensure_absolute_paths)
        cmd += other
        return cmd

    def __iadd__(self, other):
        self.append(other)
        return self

    def __iter__(self):
        """
        Support iteration so functions from :py:mod:`subprocess` module
        support `Command` instance.
        """
        return iter(self.args)

    def __repr__(self):
        return f'<{self.__class__.__name__} args={self.args}>'

    def __str__(self):
        return ' '.join(self.args)

class shellCommand(Command):
    """
        Sub Class of the command class which always executes commands as shell commands.
    """

    def __init__(self, *args, pathstyle='auto', ensure_absolute_paths=False):
        """
        :param pathstyle: Determine the path style when adding instance of
            :py:class:`pathlib.Path`. Path style determines the slash type
            which separates the path components. If pathstyle is `auto`, then
            on Windows backslashes are used and on Linux forward slashes are used.
            When backslashes should be prevented on all systems, the pathstyle
            should be `posix`. No other values are allowed.

        :param ensure_absolute_paths: If true, then any passed path will be
            converted to absolute path.

        :param args: Initial command.

        :type pathstyle: str

        :type ensure_absolute_paths: bool

        """
        return super().__init__(*args,
                                pathstyle=pathstyle,
                                ensure_absolute_paths=ensure_absolute_paths)

    def _is_shell_command(self):
        return True




import logging
import repomanager.utils as utils
import shutil
import sys
from repomanager.constants import *
from ruamel.yaml import YAML
from cerberus import Validator

yaml = YAML(typ="rt")
yaml.default_flow_style = False
yaml.allow_unicode = True
yaml.compact(seq_seq=False, seq_map=False)

logger = logging.getLogger(__name__)

schema_syn= """\
url:
    type: string
    required: True
    nullable: False
checkout:
    type: string
    required: True
    nullable: False
    default: master
recursive:
    type: boolean
    default: False
patch:
    type: list
    schema:
        type: list
        items: 
            - type: string
            - type: string
    default:
    nullable: True
sparse:
    type: list
    schema:
        type: string
    default:
    nullable: True
"""

class YamlValidator(Validator):
    def _check_with_filecheck(self, field, value):
        file_check = pwd + '/' + value
        if not os.path.isfile(file_check):
            self._error(field, 'File {0} not found'.format(file_check))
    def _check_with_dircheck(self, field, file_check):
        if not os.path.isdir(value):
            self._error(field, 'Dir {0} not found'.format(value))

def repoman(yaml_str, clean, update, patch, unpatch, work_dir):

    repo_yaml = yaml.load(yaml_str)

    schema = yaml.load(schema_syn)
    validator = YamlValidator(schema)
    validator.allow_unknown = False   

    cwd = os.getcwd()
    workdir = os.path.abspath(work_dir)
    os.makedirs(workdir, exist_ok=True)

    for name,field in repo_yaml.items():
        valid = validator.validate(field)
        if not valid:
            error_list = validator.errors
            for x in error_list:
                print('{0} [ {1} ] : {2}'.format(name,x, error_list[x]))
                exit(1)
        field = validator.normalized(field, schema)
        repo_path = workdir+'/'+name
        repo_url = field['url']
        repo_checkout = field['checkout']
        recursive = field['recursive']
        if recursive:
            recurse_cmd = '--recursive'
        else:
            recurse_cmd = ''

        if clean:
            shutil.rmtree(repo_path, ignore_errors=True, onerror=None)
            logger.info('Repo Path "{0}" Deleted'.format(name))

        if update:
            if os.path.isdir(repo_path):
                logger.info('Repo "{0}" already exists @ "{1}"'.format(name, repo_path))
                curr_commit, x = utils.shellCommand(r_currcommit).run(cwd=repo_path)
                exp_commit, x = utils.shellCommand(r_branchcommit.format(repo_checkout)).run(cwd=repo_path)
                # when we need to change commitid/branch/tags/release we first fetch everything
                # from the remote, reset the local copy and then checkout
                if curr_commit != exp_commit:
                    logger.info('Checking out "{0}" for Repo: "{1}"'.format(repo_checkout, name))
                    utils.shellCommand(r_fetch).run(cwd=repo_path) 
                    utils.shellCommand(r_reset).run(cwd=repo_path) 
                    utils.shellCommand(r_checkout.format(repo_checkout)).run(cwd=repo_path) 
                    if recursive:
                        utils.shellCommand('git submodule update --init --recursive').run(cwd=repo_path)
                else:
                    logger.info('Required Commit/Branch/Release already checked out')
            elif field['sparse'] is None:
                logger.info('Cloning "{0}" from URL "{1}"'.format(name, repo_url))
                utils.shellCommand(r_clone.format(repo_url, name, recurse_cmd)).run(cwd=workdir)
                utils.shellCommand(r_checkout.format(repo_checkout)).run(cwd=repo_path) 
            else:
                logger.info('Performing Sparse checkout')
                files = "\n".join(field['sparse'])
                cmd = r_sparseconfig.format(repo_url, files ,repo_checkout)
                os.makedirs(repo_path, exist_ok=True)
                utils.shellCommand(cmd).run(cwd=repo_path)


        if patch:
            if field['patch'] is not None:
                for p in field['patch']:
                    sm_path = os.path.join(repo_path,p[0])
                    pfile = os.path.abspath(p[1])
                    check, x = utils.shellCommand(r_checkpatch.format(pfile)).run(cwd=sm_path,
                            logging=False)
                    if check == '' and x == 0:
                        logger.info('Applying Patch "{0}" to "{0}"'.format(pfile, sm_path))
                        utils.shellCommand(r_applypatch.format(pfile)).run(cwd=sm_path, logging=False)
                    else:
                        logger.info('Patch "{0}" already applied or Cannot be applied to "{1}"'.format(pfile,sm_path))
        elif unpatch:
            if field['patch'] is not None:
                for p in field['patch']:
                    sm_path = os.path.join(repo_path,p[0])
                    pfile = os.path.abspath(p[1])
                    check, x = utils.shellCommand(r_checkunpatch.format(pfile)).run(cwd=sm_path)
                    if check == '' and x == 0:
                        logger.info('Reversing Patch "{0}" to "{0}"'.format(pfile, sm_path))
                        utils.shellCommand(r_applyunpatch.format(pfile)).run(cwd=sm_path)
                    else:
                        logger.info('Patch "{0}" already reversed or Cannot be reversed to "{1}"'.format(pfile,sm_path))






import os

root = os.path.abspath(os.path.dirname(__file__))
cwd = os.getcwd()
r_clone = 'git clone --progress {0} {1} {2};'
r_fetch = 'git fetch'
r_reset = 'git reset --hard'
r_checkout = 'git -c advice.detachedHead=false checkout {0} --recurse-submodules'
r_currcommit = 'git rev-parse HEAD'
r_branchcommit = 'git rev-parse {0}'
r_diff = 'git diff'
r_checkpatch = 'git apply --check {0}'
r_applypatch = 'git apply {0}'
r_checkunpatch = 'git apply -R --check {0}'
r_applyunpatch = 'git apply -R {0}'
r_sparseconfig = 'git init ;\
        git config core.sparseCheckout true ;\
        git remote add -f origin {0} ;\
        echo "{1}" > .git/info/sparse-checkout ;\
        git checkout {2}'
r_real_sparseconfig = 'git init ;\
        git config core.sparseCheckout true ;\
        git remote add origin {0} ;\
        echo "{1}" > .git/info/sparse-checkout ;\
        git fetch --depth 1 origin {2} ; \
        git checkout FETCH_HEAD'



import os
import codecs
from setuptools import setup, find_packages

import repomanager

# Base directory of package
here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


#Long Description
with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(name="repomanager",
    version=repomanager.__version__,
    description="Git Repository Manager",
    long_description=long_description,
    classifiers=[
          "Programming Language :: Python :: 3.6",
          "License :: OSI Approved :: BSD License",
          "Development Status :: 4 - Beta"
    ],
    url='https://gitlab.incoresemi.com/utils/repomanager',
    author='InCore Semiconductors Pvt. Ltd.',
    author_email='info@incoresemi.com',
    license='BSD-3-Clause',
    packages=find_packages(),
    keywords='repomanager',
    test_suite='tests',
    zip_safe=False,
    entry_points={
        'console_scripts': ['repomanager=repomanager.main:main'],
    })
